@@ -20,8 +20,6 @@
 /**
  * SECTION:element-dfbvideosink
  *
- * <refsect2>
- * <para>
  * DfbVideoSink renders video frames using the
  * <ulink url="http://www.directfb.org/">DirectFB</ulink> library.
  * Rendering can happen in two different modes :
@@ -47,19 +45,19 @@
  * <listitem>
  *   <para>
  *   Embedded: this mode will render video frames in a 
- *   <link linkend="GstDfbVideoSink--surface">surface</link> provided by the
+ *   #GstDfbVideoSink:surface provided by the
  *   application developer. This is a more advanced usage of the element and
  *   it is required to integrate video playback in existing 
  *   <ulink url="http://www.directfb.org/">DirectFB</ulink> applications.
  *   </para>
  *   <para>
  *   When using this mode the element just renders to the
- *   <link linkend="GstDfbVideoSink--surface">surface</link> provided by the 
+ *   #GstDfbVideoSink:surface provided by the 
  *   application, that means it won't handle navigation events and won't resize
- *   the <link linkend="GstDfbVideoSink--surface">surface</link> to fit video
+ *   the #GstDfbVideoSink:surface to fit video
  *   frames geometry. Application has to implement the necessary code to grab
  *   informations about the negotiated geometry and resize there
- *   <link linkend="GstDfbVideoSink--surface">surface</link> accordingly.
+ *   #GstDfbVideoSink:surface accordingly.
  *   </para>
  * </listitem>
  * </itemizedlist>
@@ -67,20 +65,19 @@
  * optimize memory allocation time and handle reverse negotiation. Indeed if 
  * you insert an element like videoscale in the pipeline the video sink will
  * negotiate with it to try get a scaled video for either the fullscreen layout
- * or the application provided external
- * <link linkend="GstDfbVideoSink--surface">surface</link>.
- * </para>
+ * or the application provided external #GstDfbVideoSink:surface.
+ *
+ * <refsect2>
  * <title>Example application</title>
  * <para>
  * <include xmlns="http://www.w3.org/2003/XInclude" href="element-dfb-example.xml" />
  * </para>
+ * </refsect2>
+ * <refsect2>
  * <title>Example pipelines</title>
- * <para>
- * Here is a test pipeline to test the colorbalance interface :
- * <programlisting>
+ * |[
  * gst-launch -v videotestsrc ! dfbvideosink hue=20000 saturation=40000 brightness=25000
- * </programlisting>
- * </para>
+ * ]| test the colorbalance interface implementation in dfbvideosink
  * </refsect2>
  */
 
@@ -96,6 +93,7 @@
 #include "dfbvideosink.h"
 
 #include <string.h>
+#include <stdlib.h>
 
 /* Debugging category */
 GST_DEBUG_CATEGORY_STATIC (dfbvideosink_debug);
@@ -144,6 +142,7 @@ static void gst_dfbvideosink_surface_destroy (GstDfbVideoSink * dfbvideosink,
     GstDfbSurface * surface);
 
 static GstVideoSinkClass *parent_class = NULL;
+static GstBufferClass *surface_parent_class = NULL;
 
 static const char *
 gst_dfbvideosink_get_format_name (DFBSurfacePixelFormat format)
@@ -328,18 +327,14 @@ gst_dfbvideosink_surface_destroy (GstDfbVideoSink * dfbvideosink,
     surface->surface = NULL;
   }
 
-  if (GST_BUFFER (surface)->malloc_data) {
-    g_free (GST_BUFFER (surface)->malloc_data);
-    GST_BUFFER (surface)->malloc_data = NULL;
-  }
-
   if (surface->dfbvideosink) {
     /* Release the ref to our sink */
     surface->dfbvideosink = NULL;
     gst_object_unref (dfbvideosink);
   }
 
-  return;
+  GST_MINI_OBJECT_CLASS (surface_parent_class)->finalize (GST_MINI_OBJECT
+      (surface));
 }
 
 static gpointer
@@ -584,10 +579,14 @@ gst_dfbvideosink_setup (GstDfbVideoSink * dfbvideosink)
         "DirectFB fullscreen");
     if (!dfbvideosink->dfb) {
       DFBGraphicsDeviceDescription hw_caps;
+      char *argv[] = { "-", "--dfb:quiet", NULL };
+      int argc = 2;
+      char **args;
 
       GST_DEBUG_OBJECT (dfbvideosink, "initializing DirectFB");
 
-      ret = DirectFBInit (0, NULL);
+      args = argv;
+      ret = DirectFBInit (&argc, &args);
 
       if (ret != DFB_OK) {
         GST_WARNING_OBJECT (dfbvideosink, "DirectFB initialization failed");
@@ -1540,7 +1539,7 @@ gst_dfbvideosink_show_frame (GstBaseSink * bsink, GstBuffer * buf)
 
   if (mem_cpy) {
     IDirectFBSurface *dest = NULL, *surface = NULL;
-    gpointer data;
+    guint8 *data;
     gint dest_pitch, src_pitch, line;
     GstStructure *structure;
 
@@ -1573,7 +1572,9 @@ gst_dfbvideosink_show_frame (GstBaseSink * bsink, GstBuffer * buf)
     /* Center / Clip */
     gst_video_sink_center_rect (src, dst, &result, FALSE);
 
-    res = surface->GetSubSurface (surface, (DFBRectangle *) & result, &dest);
+    res =
+        surface->GetSubSurface (surface, (DFBRectangle *) (void *) &result,
+        &dest);
     if (res != DFB_OK) {
       GST_WARNING_OBJECT (dfbvideosink, "failed when getting a sub surface");
       ret = GST_FLOW_UNEXPECTED;
@@ -1585,7 +1586,7 @@ gst_dfbvideosink_show_frame (GstBaseSink * bsink, GstBuffer * buf)
       dfbvideosink->layer->WaitForSync (dfbvideosink->layer);
     }
 
-    res = dest->Lock (dest, DSLF_WRITE, &data, &dest_pitch);
+    res = dest->Lock (dest, DSLF_WRITE, (void *) &data, &dest_pitch);
     if (res != DFB_OK) {
       GST_WARNING_OBJECT (dfbvideosink, "failed locking the external "
           "subsurface for writing");
@@ -1642,7 +1643,7 @@ gst_dfbvideosink_show_frame (GstBaseSink * bsink, GstBuffer * buf)
 
     if (dfbvideosink->hw_scaling) {
       dfbvideosink->primary->StretchBlit (dfbvideosink->primary,
-          surface->surface, NULL, (DFBRectangle *) & result);
+          surface->surface, NULL, (DFBRectangle *) (void *) &result);
     } else {
       DFBRectangle clip;
 
@@ -1891,6 +1892,8 @@ gst_dfbsurface_class_init (gpointer g_class, gpointer class_data)
 {
   GstMiniObjectClass *mini_object_class = GST_MINI_OBJECT_CLASS (g_class);
 
+  surface_parent_class = g_type_class_peek_parent (g_class);
+
   mini_object_class->finalize = (GstMiniObjectFinalizeFunction)
       gst_dfbsurface_finalize;
 }
